@@ -1,6 +1,3 @@
-import functools
-import math
-
 #     _       _
 #    (_) ___ | |_  _ __
 #    | |/ __|| __|| '__|
@@ -8,62 +5,17 @@ import math
 #    |_||___/ \__||_|
 # strings you can count on
 
-__version__ = "0.2.0"
+__version__ = "1.0.2"
 import functools
 import math
 
 """
-changelog
+Note: the changelog is now in changelog.md
 
-version 0.2.0 2024-04-30
-----------------------
-Added __iter__ method__ .
-    So now,
-        for c in istr('123'):
-            ...
-        results in c values that are istrs 
-
-Added istr.digits method:
-Examples
---------
-istr.digits() ==> istr('0123456789')
-istr.digits('') ==> istr('0123456789')
-istr.digits('1') ==> istr('1')
-istr.digits('3-') ==> istr('3456789')
-istr.digits('-3') ==> istr('0123')
-istr('1-4', '6', '8-9') ==> istr('1234689')
-istr.digits('1', '1-2', '1-3') ==> istr('112123')
-
-    Note that a digit can occur more than once.
-
-version 0.1.2  2024-04-26  
--------------------------
-Added all relevant string methods to return istrs or data structures with istrs.
-Added corresponding tests.
-
-version 0.1.0  2024-04-22  
--------------------------
-Changed the way istr.range is implemennted.
-
-Changed the context manager istr.format() to be used directly without the with statement.
-Also, noww istr.format() works without any argument and then returns the current format.
-
-istr class now uses __slots__
-
-All internal values and methods now start with an underscore.
-
-Introduced istr.repr_mode()
-
-Introduced istr.base()
-
-Extended tests for new functionality
-
-
-version 0.0.8  2024-04-18  
--------------------------
-initial version with changelog
+You can view the changelog on www.salabim/istr_changelog.html
 """
 
+_0_to_Z = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
 
 class _range:
     """
@@ -241,9 +193,12 @@ class istr(str):
 
     __slots__ = ("_as_int", "_as_repr")
 
-    _format = ""
-    _mode = "istr"
+    _int_format = ""
+    _repr_mode = "istr"
     _base = 10
+    _nan = object()
+    _force_istr_repr = False
+    _digits_cache = {}
 
     @staticmethod
     def _to_base(number, base):
@@ -251,7 +206,7 @@ class istr(str):
             raise ValueError(f"negative numbers are not allowed for base {base}")
         result = ""
         while number:
-            result += "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"[number % base]
+            result += _0_to_Z[number % base]
             number //= base
         return result[::-1] or "0"
 
@@ -262,8 +217,30 @@ class istr(str):
                 return int(value, cls._base)
             else:
                 return int(value)
-        except (ValueError, TypeError):
-            raise ValueError(f"unable to convert {repr(value)} to int (base {cls._base})")
+        except:
+            return cls._nan
+
+    def _check_is_int(self, *args, right=False):
+        operator = args[len(args) == 2]
+        if len(args) == 2:
+            other = args[0]
+            if not self.is_int() or self._to_int(other) is self._nan:
+                self.__class__._force_istr_repr = True
+                if right:
+                    message = f"unsupported operand for {operator}: {repr(other)} and {repr(self)}"
+                else:
+                    message = f"unsupported operand for {operator}: {repr(self)} and {repr(other)}"
+                self.__class__._force_istr_repr = False
+                raise TypeError(message)
+        else:
+            if not self.is_int():
+                self.__class__._force_istr_repr = True
+                message = f"unsupported operand for {operator}: {repr(self)}"
+                self.__class__._force_istr_repr = False
+                raise TypeError(message)
+
+    def _rcheck_is_int(self, *args):
+        self._check_is_int(*args, right=True)
 
     def __new__(cls, *value):
         if len(value) == 0:
@@ -282,32 +259,31 @@ class istr(str):
             if hasattr(value, "__next__"):
                 return map(functools.partial(cls), value)
             return type(value)(map(functools.partial(cls), value))
-        if value == "":
-            as_str = ""
-            as_int = 0
+
+        as_int = cls._to_int(value)
+        if isinstance(value, str):
+            as_str = value
         else:
-            as_int = cls._to_int(value)
-            if (cls._format == "" or cls._base != 10) and not isinstance(value, istr):
-                if isinstance(value, str):
-                    as_str = value
+            if as_int is cls._nan:
+                raise TypeError(f"incorrect value for {cls.__name__}: {repr(value)}")
+            if cls._int_format == "" or cls._base != 10:
+                if cls._base == 10:
+                    as_str = str(as_int)
                 else:
-                    if cls._base == 10:
-                        as_str = str(as_int)
-                    else:
-                        as_str = istr._to_base(as_int, cls._base)
+                    as_str = istr._to_base(as_int, cls._base)
             else:
-                as_str = f"{as_int:{cls._format}}"
+                as_str = f"{as_int:{cls._int_format}}"
 
         self = super().__new__(cls, as_str)
         self._as_int = as_int
-        if self._mode == "istr":
+        if self._repr_mode == "istr":
             self._as_repr = f"{cls.__name__}({repr(as_str)})"
-        elif self._mode == "int":
-            self._as_repr = repr(as_int)
+        elif self._repr_mode == "int":
+            self._as_repr = "?" if as_int is self._nan else repr(as_int)
         else:
             self._as_repr = repr(as_str)
         return self
-              
+
     def __iter__(self):
         yield from self.__class__(super().__iter__())
 
@@ -316,7 +292,8 @@ class istr(str):
 
     def __eq__(self, other):
         if isinstance(other, self.__class__):
-            return self._as_int == other._as_int
+            if self.is_int() and other.is_int():
+                return self._as_int == other._as_int
         if isinstance(other, str):
             return super().__eq__(other)
         try:
@@ -331,63 +308,85 @@ class istr(str):
         return super().__contains__(str(other))
 
     def __repr__(self):
+        if self._force_istr_repr:
+            return f"{self.__class__.__name__}({super().__repr__()})"
         return self._as_repr
 
     def __le__(self, other):
+        self._check_is_int(other, "<=")
         return self._as_int <= self._to_int(other)
 
     def __lt__(self, other):
+        self._check_is_int(other, "<")
         return self._as_int < self._to_int(other)
 
     def __ge__(self, other):
+        self._check_is_int(other, ">=")
         return self._as_int >= self._to_int(other)
 
     def __gt__(self, other):
+        self._check_is_int(other, ">")
         return self._as_int > self._to_int(other)
 
     def __bool__(self):
-        return bool(self._as_int)
+        if self.is_int():
+            return bool(self._as_int)
+        return bool(str(self))
 
     def __add__(self, other):
+        self._check_is_int(other, "+")
         return self.__class__(self._as_int + self._to_int(other))
 
+    def __radd__(self, other):
+        self._rcheck_is_int(other, "+")
+        return self.__class__(self._to_int(other) + self._as_int)
+
     def __sub__(self, other):
+        self._check_is_int(other, "-")
         return self.__class__(self._as_int - self._to_int(other))
 
+    def __rsub__(self, other):
+        self._rcheck_is_int(other, "-")
+        return self.__class__(self._to_int(other) - self._as_int)
+
     def __mul__(self, other):
+        self._check_is_int(other, "*")
         return self.__class__(self._as_int * self._to_int(other))
 
+    def __rmul__(self, other):
+        self._rcheck_is_int(other, "*")
+        return self.__class__(self._to_int(other) * self._as_int)
+
     def __floordiv__(self, other):
+        self._check_is_int(other, "//")
         return self.__class__(self._as_int // self._to_int(other))
 
     def __rfloordiv__(self, other):
+        self._rcheck_is_int(other, "//")
         return self.__class__(self._to_int(other) // self._as_int)
 
     def __truediv__(self, other):
+        self._check_is_int(other, "/")
         return self.__class__(self._as_int // self._to_int(other))
 
     def __rtruediv__(self, other):
+        self._rcheck_is_int(other, "/")
         return self.__class__(self._to_int(other) // self._as_int)
 
     def __pow__(self, other):
+        self._check_is_int(other, "**")
         return self.__class__(self._as_int ** self._to_int(other))
 
     def __rpow__(self, other):
+        self._rcheck_is_int(other, "**")
         return self.__class__(self._to_int(other) ** self._as_int)
 
-    def __radd__(self, other):
-        return self.__class__(self._to_int(other) + self._as_int)
-
-    def __rsub__(self, other):
-        return self.__class__(self._to_int(other) - self._as_int)
-
-    def __rmul__(self, other):
-        return self.__class__(self._to_int(other) * self._as_int)
-
     def __mod__(self, other):
+        self._check_is_int(other, "%")
         return self.__class__(self._as_int % self._to_int(other))
 
     def __rmod__(self, other):
+        self._rcheck_is_int(other, "%")
         return self.__class__(self._to_int(other) % self._as_int)
 
     def __or__(self, other):
@@ -396,19 +395,25 @@ class istr(str):
     def __ror__(self, other):
         return self.__class__("".join((self.__class__(other), self)))
 
-    def __int__(self):
-        return int(self._as_int)
-
     def __round__(self):
+        self._check_is_int("round")
         return self.__class__(round(self._as_int))
 
+    def __int__(self):
+        if not self.is_int():
+            raise ValueError(f"invalid literal for int() with base 10: {repr(self)}")
+        return int(self._as_int)
+
     def __trunc__(self):
+        self._check_is_int("trunc")
         return self.__class__(math.trunc(self._as_int))
 
     def __floor__(self):
+        self._check_is_int("floor")
         return self.__class__(math.floor(self._as_int))
 
     def __ceil__(self):
+        self._check_is_int("ceil")
         return self.__class__(math.ceil(self._as_int))
 
     def __matmul__(self, other):
@@ -418,26 +423,39 @@ class istr(str):
         return self.__class__(super().__rmul__(other))
 
     def __divmod__(self, other):
+        self._check_is_int(other, "divmod")
         return self.__class__(divmod(self._as_int, self._to_int(other)))
 
     def __rdivmod__(self, other):
+        self._rcheck_is_int(other, "divmod")
         return self.__class__(divmod(self._to_int(other), self._as_int))
 
     def __neg__(self):
+        self._check_is_int("-")
         return self.__class__(-self._as_int)
 
     def __pos__(self):
+        self._check_is_int("+")
         return self
 
     def __abs__(self):
+        self._check_is_int("abs")
         return self.__class__(abs(self._as_int))
 
     def is_even(self):
+        self._check_is_int("is_even")
         return self._as_int % 2 == 0
 
     def is_odd(self):
+        self._check_is_int("is_odd")
         return self._as_int % 2 == 1
 
+    def all_distinct(self):
+        return len(self) == len(set(self))
+
+    def is_int(self):
+        return self._as_int is not self._nan
+
     def reversed(self):
         return self[::-1]
 
@@ -454,45 +472,45 @@ class istr(str):
             yield cls(i), value
 
     @classmethod
-    class format:
-        def __new__(cls, cls_format, format=None):
-            if format is None:
-                return cls_format._format
+    class int_format:
+        def __new__(cls, cls_int_format, int_format=None):
+            if int_format is None:
+                return cls_int_format._int_format
             return super().__new__(cls)
 
-        def __init__(self, cls, format):
-            self.saved_format = cls._format
+        def __init__(self, cls, int_format):
+            self.saved_int_format = cls._int_format
             self.saved_cls = cls
-            if not (isinstance(format, str) and all(x in "0123456789" for x in format)):
-                raise ValueError(f"{repr(format)} is incorrect format")
+            if not (isinstance(int_format, str) and all(x in "0123456789" for x in int_format)):
+                raise ValueError(f"{repr(int_format)} is incorrect int_format")
 
-            cls._format = format
+            cls._int_format = int_format
 
         def __enter__(self):
             ...
 
         def __exit__(self, exc_type, exc_value, exc_tb):
-            self.saved_cls._format = self.saved_format
+            self.saved_cls._int_format = self.saved_int_format
 
     @classmethod
     class repr_mode:
-        def __new__(cls, cls_mode, mode=None):
+        def __new__(cls, cls_repr_mode, mode=None):
             if mode is None:
-                return cls_mode._mode
-            if mode in ("istr", "str", "int"):
+                return cls_repr_mode._repr_mode
+            if mode in ("istr", "str", "int"):  # _istr is used only for TypeErrors
                 return super().__new__(cls)
             raise TypeError(f"mode not 'istr', 'str' or 'int', but {repr(mode)}")
 
         def __init__(self, cls, mode):
-            self.saved_mode = cls._mode
+            self.saved_repr_mode = cls._repr_mode
             self.saved_cls = cls
-            cls._mode = mode
+            cls._repr_mode = mode
 
         def __enter__(self):
             ...
 
         def __exit__(self, exc_type, exc_value, exc_tb):
-            self.saved_cls._mode = self.saved_mode
+            self.saved_cls._repr_mode = self.saved_repr_mode
 
     @classmethod
     class base:
@@ -517,19 +535,21 @@ class istr(str):
     @classmethod
     def range(cls, start, stop=None, step=1):
         return _range(cls, start, stop, step)
-        
+
     @classmethod
-    @functools.lru_cache
-    def digits(cls,*args):
-        '''
+    def digits(cls, *args):
+        """
         return an istr of istr'ed digits as specified with args
-        
+
         if no args, 0-9 will be used
-        
+
         all given args will be used
         each arg has to be either null string, <digit>, <digit>-<digit> or -<digit>
-        
-        examples
+
+        the digits may be '0' through '9' and 'A' through 'Z' (not case sensitive)
+        The returned value will always be in uppercase (if applicable).
+
+        Examples
         --------
         istr.digits() ==> istr('0123456789')
         istr.digits('') ==> istr('0123456789')
@@ -538,42 +558,52 @@ class istr(str):
         istr.digits('-3') ==> istr('0123')
         istr('1-4', '6', '8-9') ==> istr('1234689')
         istr('1', '1-2', '1-3') ==> istr('11213')
+        istr.digits('-z') ==> istr('0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ')
+        istr.digits('C') ==> istr('C')
+        istr.digits('A-F') ==> istr('ABCDEF')
+        istr.digits('X-') ==> istr('XYZ')
 
         Note
         ----
         A digit can occur more than once.
-
-        '''
-        result=[]
+        """
+        key = (args, cls._base, cls._int_format, cls._repr_mode)
+        if key in cls._digits_cache:
+            return cls.digits_cache[key]
+        result = []
         if not args:
-            args=['0-9']
+            args = ["0-9"]
         for arg in args:
-            if arg.strip()=='':
-                arg='0-9'
-            pre,*post=arg.split('-')
-            if pre.strip()=='':
-                pre='0'
-            try:
-                start=int(pre)
-            except ValueError:
-                raise ValueError(f'incorrect specifier: {repr(arg)}')  
-            if not 0<=start<=9:
-                raise ValueError(f'incorrect specifier: {repr(arg)}')                
-            if len(post)>1:
-                raise ValueError(f'incorrect specifier: {repr(arg)}')
+            if arg.strip() == "":
+                arg = "0-9"
+            pre, *post = arg.split("-", 1)
+            if pre.strip() == "":
+                pre = "0"
+            pre = pre.upper()
+            if len(pre) > 1 or pre not in _0_to_Z:
+                raise ValueError(f"incorrect specifier: {repr(arg)}")
+            start = _0_to_Z.index(pre)
+
             if post:
-                if post[0].strip()=="":
-                    post="9"
-                try:
-                    stop=int(post[0])
-                except ValueError:
-                    raise ValueError(f'incorrect specifier: {repr(arg)}')                    
-                if (not 0<=start<=9) or start>stop:
-                    raise ValueError(f'incorrect specifier: {repr(arg)}')
+                post = post[0]
+                if post.strip() == "":
+                    if pre in "0123456789":
+                        post = "9"
+                    else:
+                        post = "Z"
+                post = post.upper()
+                if len(post) > 1 or post not in _0_to_Z:
+                    raise ValueError(f"incorrect specifier: {repr(arg)}")
+                stop = _0_to_Z.index(post)
+                if start > stop:
+                    raise ValueError(f"incorrect specifier: {repr(arg)}")
             else:
-                stop=start
-            result.extend(range(start,stop+1))
-        return cls('').join(istr(result))
+                stop = start
+            result.extend(_0_to_Z[i] for i in range(start, stop + 1))
+
+        result = istr("".join(result))
+        cls._digits_cache[key] = result
+        return result
 
     def capitalize(self, *args, **kwargs):
         return self.__class__(super().capitalize(*args, **kwargs))
@@ -587,6 +617,9 @@ class istr(str):
     def expandtabs(self, *args, **kwargs):
         return self.__class__(super().expandtabs(*args, **kwargs))
 
+    def format(self, *args, **kwargs):
+        return self.__class__(super().format(*args, **kwargs))
+
     def join(self, *args, **kwargs):
         return self.__class__(super().join(*args, **kwargs))
 
@@ -646,15 +679,15 @@ class istr(str):
 
 
 def main():
-    print(repr(istr.digits()))
-    print(istr.digits('1-9'))
-    print(istr.digits('1'))
-    print(istr.digits('1-2', '6'))    
-    print(istr.digits('1-5', '7-9'))        
-    print(istr.digits('-5'))   
-    print(istr.digits('3-'))   
-    print(repr(divmod(istr(20),3)))
+    with istr.repr_mode("int"):
+        print(repr(istr(3)))
+        print(repr(istr("a")))
+        with istr.int_format("4"):
+            print(repr(istr(3)))
+            print(repr(istr("a")))
+        with istr.int_format("04"):
+            print(repr(istr(3)))
+            print(repr(istr("a")))
 
 if __name__ == "__main__":
     main()
-
